#include <stdio.h>
#include <errno.h>

#include "fail.h"
#include "conf.h"
#include "ttybuf.h"
#include "remtty.h"

#define MAXLINE 127

typedef struct hostentry {
  struct hostentry *next;
  char *host;
  int port;
} hostentry;

static hostentry *firsthost = NULL;

static void addhost(char *line) {
  hostentry *lasthost;
  hostentry *h = (hostentry *) malloc(sizeof(hostentry));
  if (h == NULL) {
    fail("Unable to allocate host entry.\n");
  }

  h->next = NULL;
  h->host = (char *) malloc(strlen(line)+1);

  if (2 != sscanf(line, "%[^:]:%d", h->host, &h->port)) {
    h->port = 23;
  }

  if (firsthost == NULL) {
    firsthost = h;
  } else {
    lasthost = firsthost;
    while (lasthost->next != NULL) {
      lasthost = lasthost->next;
    }
    lasthost->next = h;
  }
}

typedef struct chatline {
  struct chatline *next;
  char *expect;
  char *send;
} chatline;

static chatline *firstchatline = NULL;

static void translatemeta(char *src, char *dest) {
  int i = 0;
  int j = 0;
  while (src[i] != '\0') {
    if (src[i] == '\\') {
      if (src[i+1] == 'r') {
	dest[j++] = '\r'; i += 2;
      } else if (src[i+1] == 'n') {
	dest[j++] = '\n'; i += 2;
      } else {
	dest[j++] = src[i++];
      }
    } else {
      dest[j++] = src[i++];
    }
  }
  dest[j] = '\0';
}

static void addchatline(char *line) {
  char *tmp;
  chatline *lastchatline;
  chatline *c = (chatline *) malloc(sizeof(chatline));
  if (c == NULL) {
    fail("Unable to allocate chat line entry.\n");
  }

  c->next = NULL;
  c->expect = (char *) malloc(strlen(line)+1);
  c->send = (char *) malloc(strlen(line)+1);

  if (2 != sscanf(line, "\"%[^\"]\",\"%[^\"]\"", c->expect, c->send)) {
    fail("Invalid format of line: %s.\n", line);
  }

  tmp = (char *) malloc(strlen(c->expect));
  translatemeta(c->expect, tmp);
  strcpy(c->expect, tmp);
  free(tmp);
  
  tmp = (char *) malloc(strlen(c->send));
  translatemeta(c->send, tmp);
  strcpy(c->send, tmp);
  free(tmp);

  if (firstchatline == NULL) {
    firstchatline = c;
  } else {
    lastchatline = firstchatline;
    while (lastchatline->next != NULL) {
      lastchatline = lastchatline->next;
    }
    lastchatline->next = c;
  }
}

void config_load(char *cfg) {
  FILE *config;
  char line[MAXLINE+1];

  config = fopen(cfg, "r");
  if (NULL == config) {
    fail("Error while opening configuration file: %s.\n", cfg);
  }

  while (NULL != fgets(line, MAXLINE, config)) {
    char *token;
    char *nexttoken;
    char *comment;

    comment = (char *) strrchr(line, '#');
    if (comment != NULL) {
      *comment = '\0';		/* Clear the comment */
    }

    nexttoken = line;
    token = (char *) strsep(&nexttoken, " \t");
    if (NULL != token) {
      if (0 == strcmp(token, "host:")) {
	addhost(nexttoken);
      } else if (0 == strcmp(token, "chat:")) {
	addchatline(nexttoken);
      } else {
	fail("Unknown token in configuration file: %s.\n", token);
      }
    }
  }
  
  fclose(config);
}

static int logged = 0;

int config_logged() {
  return logged;
}

int config_chat(unsigned char c) {
  static chatline *current = NULL;
  static int w = 0;

  if (current == NULL) {
    current = firstchatline;
  }

  if (current->expect[w] == c) {
    w++;
    if (verbose) tty_put1(c);
    if (w == strlen(current->expect)) {
      sock_put(current->send, strlen(current->send));
      if (current->next == NULL) {
	sock_sethandler(tty_put1);
	logged = 1;
      } else {
	current = current->next;
      }
      return 0;
    }
  } else {
    w = 0;
    if (verbose) tty_put1(c);
  }
  return 1;
}

int config_gethost(char **host, int *port) {
  hostentry *h = firsthost;

  if (h == NULL) {
    return 0;
  }

  *host = h->host;
  *port = h->port;

  firsthost = h->next;
  return 1;
}
