#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <pty.h>
#include <errno.h>
#include <sys/time.h>
#include <getopt.h>

#include "ttybuf.h"
#include "sockbuf.h"
#include "sock.h"
#include "stty.h"
#include "conf.h"
#include "pty.h"
#include "remtty.h"

int verbose = 0;
char version[] = "remtty, version " VERSION;
char *progname;
char usage[] = "%s [-hvVcx] [--help] [--version] [--verbose] [--config remtty.conf] [--comm commprog]\n";
char default_config[] = SYSCONFDIR "/remtty.conf";

void login() {
  fd_set rfds, wfds;

  sockr_reset();
  sockw_reset();
  ttyr_reset();
  ttyw_reset();

  sock_sethandler(config_chat);
  tty_sethandler(sock_put1);

  while (!config_logged() && sock_alive()) {
    int maxfd = 0;

    FD_ZERO(&rfds);
    FD_ZERO(&wfds);

    if (ttyw_ready()) {
      maxfd = sock_fd();
      FD_SET(sock_fd(), &rfds);
    }
    if (sockw_has_data()) {
      maxfd = sock_fd();
      FD_SET(sock_fd(), &wfds);
    }
    if (sockw_ready()) {
      maxfd = (maxfd > tty_rfd()) ? maxfd : tty_rfd();
      FD_SET(tty_rfd(), &rfds);
    }
    if (ttyw_has_data()) {
      maxfd = (maxfd > tty_wfd()) ? maxfd : tty_wfd();
      FD_SET(tty_wfd(), &wfds);
    }

    if (select(maxfd + 1, &rfds, &wfds, NULL, NULL) < 0) {
      if (errno != EINTR) {
	warn("Error while select(): %s.\n", strerror(errno));
	continue;
      }
    }

    if (FD_ISSET(sock_fd(), &wfds)) { sock_write();    }
    if (FD_ISSET(tty_wfd(), &wfds)) { tty_write();     }
    if (FD_ISSET(sock_fd(), &rfds)) { sock_readloop(); }
    if (FD_ISSET(tty_rfd(), &rfds)) { tty_readloop();  }
  }

  sock_flush();
  tty_flush();
}

void mainloop() {
  fd_set rfds, wfds;

  sockr_reset();
  sockw_reset();
  ttyr_reset();
  ttyw_reset();

  sock_sethandler(tty_put1);
  tty_sethandler(sock_put1);

  while (sock_alive()) {
    int maxfd = 0;

    FD_ZERO(&rfds);
    FD_ZERO(&wfds);

    if (ttyw_ready()) {
      maxfd = sock_fd();
      FD_SET(sock_fd(), &rfds);
    }
    if (sockw_has_data()) {
      maxfd = sock_fd();
      FD_SET(sock_fd(), &wfds);
    }
    if (sockw_ready()) {
      maxfd = (maxfd > tty_rfd()) ? maxfd : tty_rfd();
      FD_SET(tty_rfd(), &rfds);
    }
    if (ttyw_has_data()) {
      maxfd = (maxfd > tty_wfd()) ? maxfd : tty_wfd();
      FD_SET(tty_wfd(), &wfds);
    }

    if (select(maxfd + 1, &rfds, &wfds, NULL, NULL) < 0) {
      if (errno != EINTR) {
	warn("Error while select(): %s.\n", strerror(errno));
	continue;
      }
    }

    if (FD_ISSET(sock_fd(), &wfds)) { sock_write();    }
    if (FD_ISSET(tty_wfd(), &wfds)) { tty_write();     }
    if (FD_ISSET(sock_fd(), &rfds)) { sock_readloop(); }
    if (FD_ISSET(tty_rfd(), &rfds)) { tty_readloop();  }
  }
}

void help() {
  fputs("Lets you access a network connection with a pseudo tty.\n", stderr);
  fprintf(stderr, usage, progname);
  fputs("  -h, --help\t\tPrint this help message\n", stderr);
  fputs("  -v, --verbose\t\tSwitch to verbose mode\n", stderr);
  fputs("  -V, --version\t\tPrint the version number\n", stderr);
  fprintf(stderr, "  -c, --config\t\tSpecify the configuration file (default %s)\n", default_config);
  fputs("  -x, --comm\t\tCommunications program to run, %p is replaced by the pty\n", stderr);
  exit(EXIT_SUCCESS);
}

static struct option long_options[] = {
  { "config",  1, 0, 'c' },
  { "verbose", 0, 0, 'v' },
  { "help",    0, 0, 'h' },
  { "version", 0, 0, 'V' },
  { "comm",    0, 0, 'x' },
  { 0, 0, 0, 0}
};

int main(int argc, char *argv[]) {
  char *commprog = NULL;
  char *config = default_config;
  char *pty;
  int optc;

  progname = argv[0];

  while ((optc = getopt_long(argc, argv, "c:hvVx:", long_options, (int *) 0)) != EOF) {
    switch (optc) {
    case 'c':
      config = optarg;
      break;
    case 'h':
      help();
      break;
    case 'v':
      verbose = 1;
      break;
    case 'V':
      fprintf(stderr, "%s\n", version);
      exit(EXIT_SUCCESS);
    case 'x':
      commprog = optarg;
      break;
    default:
      fprintf(stderr, usage, progname);
      exit(EXIT_FAILURE);
    }
  }
  
  config_load(config);

  while (1) {
    char *hostname;
    int port;

    if (config_gethost(&hostname, &port)) {
      info("Trying to connect to %s:%d\n", hostname, port);
      if (sock_open(hostname, port)) {
	info("Connected to %s:%d\n", hostname, port);
	break;
      }
    } else {
      fail("Unable to find a host to connect to.\n");
    }
  }

  if (commprog != NULL) {
    int fd = get_pty(&pty);
    info("Allocated pty: %s (fd=%d).\n", pty, fd);
    if (fd == -1) {
      fail("No pty could be allocated.\n");
    }
    tty_set(fd, fd);
  } else {
    info("Using fds 0 and 1.\n");
    tty_set(0, 1);
    set_tty();
  }

  login();

  if (commprog != NULL) {
    forkcomm(commprog, pty);
  }

  mainloop();

  exit(EXIT_SUCCESS);
}
